"""Tools for controlling atomic force microscopes.

Provides control of AFM postition using both short-range (piezo) and
long range (stepper) vertical positioning.  There are separate modules
for controlling the piezo (`pypiezo`) and stepper (`stepper`), this
module only contains methods that require the capabilities of both.
"""

import logging as _logging

from pypiezo.base import convert_volts_to_bits as _convert_volts_to_bits
from pypiezo.base import convert_meters_to_bits as _convert_meters_to_bits
from pypiezo.base import convert_bits_to_meters as _convert_bits_to_meters
from pypiezo.surface import SurfaceError as _SurfaceError
from pypiezo.surface import FlatFit as _FlatFit


__version__ = '0.1'


LOG = _logging.getLogger('pyafm')
"pyafm logger"

LOG.setLevel(_logging.DEBUG)
h = _logging.StreamHandler()
h.setLevel(_logging.WARN)
f = _logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
h.setFormatter(f)
LOG.addHandler(h)
del h, f


class AFM (object):
    """Atomic force microscope positioning.

    Uses a short range `piezo` and a long range `stepper` to position
    an AFM tip relative to the surface.

    Parameters
    ----------
    piezo | pypiezo.afm.AFMpiezo instance
        Fine positioning and deflection measurements.
    stepper | stepper.Stepper instance
        Coarse positioning.
    """
    def __init__(self, piezo, stepper, axis_name='z'):
        self.piezo = piezo
        self.stepper = stepper
        self.axis_name = axis_name

    def move_just_onto_surface(self, depth=-50e-9, setpoint=2, far=200):
        """Position the AFM tip close to the surface.

        Uses `.piezo.get_surface_position()` to pinpoint the position
        of the surface.  Adjusts the stepper position as required via
        `.stepper.step_relative()` to get within
        `2*.stepper.step_size` meters of the surface.  Then adjusts
        the piezo to place the cantilever `depth` meters onto the
        surface.  Negative `depth`\s place the tip off the surface

        If `.piezo.get_surface_position()` fails to find the surface,
        backs off `far` half steps (for safety) and steps in (without
        moving the zpiezo) until deflection voltage is greater than
        `setpoint`.
        """
        LOG.info('moving to %g onto the surface' % depth)

        stepper_tolerance = 2*self.stepper.step_size

        axis = self.piezo.axis_by_name(self.axis_name)

        zero = _convert_volts_to_bits(axis.axis_channel_config, 0)
        target_def = _convert_volts_to_bits(axis.axis_channel_config, setpoint)

        LOG.debug('zero the %s piezo output' % self.axis_name)
        self.piezo.jump(axis_name=self.axis_name, position=zero)

        LOG.debug("see if we're starting near the surface")
        try:
            pos = self.piezo.get_surface_position(
                axis_name=self.axis_name, max_deflection=target_def)
        except _FlatFit, e:
            LOG.info(e)
            pos = self._stepper_approach_again(
                target_deflection=target_def, far=far)
        except _SurfaceError, e:
            LOG.info(e)
            pos = self._stepper_approach_again(
                target_deflection=target_def, far=far)

        pos_m = _convert_bits_to_meters(
            axis.axis_channel_config, axis.axis_config, pos)
        LOG.debug('located surface at stepper %d, piezo %d (%g m)'
                  % (self.stepper.position, pos, pos_m))

        LOG.debug('fine tune the stepper position')
        while pos_m < -stepper_tolerance:  # step back if we need to
            LOG.debug('step back')
            self.stepper.step_relative(-1, backlash_safe=True)
            pos = self.piezo.get_surface_position(
                axis_name=self.axis_name, max_deflection=target_def)
            pos_m = _convert_bits_to_meters(
                axis.axis_channel_config, axis.axis_config, pos)
            LOG.debug('located surface at stepper %d, piezo %d (%g m)'
                      % (self.stepper.position, pos, pos_m))
        while pos_m > stepper_tolerance:  # step forward if we need to
            LOG.debug('step forward')
            self.stepper.step_relative(1)
            pos = self.piezo.get_surface_position(
                axis_name=self.axis_name, max_deflection=target_def)
            pos_m = _convert_bits_to_meters(
                axis.axis_channel_config, axis.axis_config, pos)
            LOG.debug('located surface at stepper %d, piezo %d (%g m)'
                      % (self.stepper.position, pos, pos_m))

        LOG.debug('adjust the %s piezo to place us just onto the surface'
                  % self.axis_name)
        target_m = pos_m + depth
        target = _convert_meters_to_bits(
            axis.axis_channel_config, axis.axis_config, target_m)
        self.piezo.jump(self.axis_name, target)

        LOG.debug(
            'positioned %g m into the surface at stepper %d, piezo %d (%g m)'
            % (depth, self.stepper.position, target, target_m))


    def _stepper_approach_again(target_deflection, far):
        LOG.info('back off %d half steps and approach until deflection > %g'
                 % (far, target_deflection))

        # back away
        self.stepper.step_relative(-far, backlash_safe=True)

        cd = self.pyiezo.read_deflection()  # cd = current deflection in bits
        LOG.debug('single stepping approach')
        while cd < target_deflection:
            LOG.debug('deflection %g < setpoint %g.  step closer'
                      % (cd, target_deflection))
            self.stepper.single_step(1)  # step in
            cd = self.pyiezo.read_deflection()

        for i in range(2*max(1, self.stepper.backlash)):
            LOG.debug('additional surface location attempt')
            try:
                pos = self.piezo.get_surface_position(
                    axis_name=self.axis_name, max_deflection=target_def)
                return pos
            except _SurfaceError, e:
                LOG.info(e)
            self.stepper.single_step(-1)  # step out
        LOG.debug('giving up on finding the surface')
        LOG.warn(e)
        raise e
